##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Constant definitions for built-in Zope permissions
"""

access_contents_information = 'Access contents information'
add_database_methods = 'Add Database Methods'
add_documents_images_and_files = 'Add Documents, Images, and Files'
add_external_methods = 'Add External Methods'
add_folders = 'Add Folders'
add_mailhost_objects = 'Add MailHost objects'
add_page_templates = 'Add Page Templates'
add_python_scripts = 'Add Python Scripts'
add_user_folders = 'Add User Folders'
add_vocabularies = 'Add Vocabularies'
add_z_gadfly_database_connections = 'Add Z Gadfly Database Connections'
add_zcatalogs = 'Add ZCatalogs'
add_zope_tutorials = 'Add Zope Tutorials'
change_database_connections = 'Change Database Connections'
change_database_methods = 'Change Database Methods'
change_external_methods = 'Change External Methods'
change_images_and_files = 'Change Images and Files'
change_python_scripts = 'Change Python Scripts'
change_configuration = 'Change configuration'
change_page_templates = 'Change Page Templates'
change_permissions = 'Change permissions'
change_proxy_roles = 'Change proxy roles'
copy_or_move = 'Copy or Move'
create_class_instances = 'Create class instances'
define_permissions = 'Define permissions'
delete_objects = 'Delete objects'
edit_factories = 'Edit Factories'
ftp_access = 'FTP access'
import_export_objects = 'Import/Export objects'
manage_vocabulary = 'Manage Vocabulary'
manage_zcatalog_entries = 'Manage ZCatalog Entries'
manage_zcatalog_indexes = 'Manage ZCatalogIndex Entries'
manage_properties = 'Manage properties'
manage_users = 'Manage users'
open_close_database_connection = 'Open/Close Database Connection'
open_close_database_connections = 'Open/Close Database Connections'
query_vocabulary = 'Query Vocabulary'
search_zcatalog = 'Search ZCatalog'
take_ownership = 'Take ownership'
test_database_connections = 'Test Database Connections'
undo_changes = 'Undo changes'
use_database_methods = 'Use Database Methods'
use_factories = 'Use Factories'
use_mailhost_services = 'Use mailhost services'
view = 'View'
view_management_screens = 'View management screens'
webdav_access = 'WebDAV access'
webdav_lock_items = 'WebDAV Lock items'
webdav_manage_locks = 'Manage WebDAV Locks'
webdav_unlock_items = 'WebDAV Unlock items'
